"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Design = void 0;
var resolver_1 = require("./resolver");
var toResource = function (resolvable, finalize) {
    if (finalize === void 0) { finalize = function () { return Promise.resolve(); }; }
    return ({
        resolve: function (injector) { return __awaiter(void 0, void 0, void 0, function () { return __generator(this, function (_a) {
            return [2 /*return*/, resolvable(injector)];
        }); }); },
        finalize: finalize,
    });
};
/**
 * Design represents key-value styled dependency graph which can detect which dependent key is missing at compile time.
 *
 * i.g.)
 * ```
 * type HasNumber = { numberKey: number }
 *
 * // NOTE: `(injector: Injector<T>) => Promise<U>` represents dependencies for a value U.
 * // Injector<T> resolves value T[keyof T] by calling `injector.key()` asynchronously.
 * const detectStringFromNumber =
 *   async (injector: Injector<HasNumber>) => `number is ${await injector.numberKey}`;
 *
 * const design = Design.empty
 *   .bind('stringFromNumber', detectStringFromNumber);
 *
 * design.resolve({}) // compile error
 * design.resolve({ numberKey: 123 }) // returns Promise({ numberKey: 123, stringFromNumber: 'number is 123' })
 * ```
 *
 * see specs for more examples!
 *
 */
var Design = /** @class */ (function () {
    function Design(design) {
        var _this = this;
        this.design = design;
        this.bind = function (key, resolvable, finalize) {
            var _a;
            if (finalize === void 0) { finalize = function () { return Promise.resolve(); }; }
            var underlying = __assign(__assign({}, _this.design), (_a = {}, _a[key] = toResource(resolvable, finalize), _a));
            return new Design(underlying);
        };
        this.bindResource = function (key, resolvable) {
            return _this.bind(key, resolvable, function (resource) { return resource.finalize(); });
        };
        this.merge = function (that) {
            var underlying = __assign(__assign({}, _this.design), that.design);
            return new Design(underlying);
        };
        this.resolve = function (requirements) {
            return resolver_1.resolve(_this.merge(Design.pure(requirements)).design);
        };
        this.use = function (requirements) { return function (f) { return __awaiter(_this, void 0, void 0, function () {
            var _a, container, finalize, result;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0: return [4 /*yield*/, this.resolve(requirements)];
                    case 1:
                        _a = _b.sent(), container = _a.container, finalize = _a.finalize;
                        _b.label = 2;
                    case 2:
                        _b.trys.push([2, , 4, 6]);
                        return [4 /*yield*/, f(container)];
                    case 3:
                        result = _b.sent();
                        return [2 /*return*/, result];
                    case 4: return [4 /*yield*/, finalize()];
                    case 5:
                        _b.sent();
                        return [7 /*endfinally*/];
                    case 6: return [2 /*return*/];
                }
            });
        }); }; };
    }
    Design.pure = function (mapping) {
        var design = {};
        var _loop_1 = function (key) {
            design[key] = {
                resolve: function () { return Promise.resolve(mapping[key]); },
                finalize: function () { return Promise.resolve(); },
            };
        };
        for (var key in mapping) {
            _loop_1(key);
        }
        return new Design(design);
    };
    Design.empty = new Design({});
    Design.bind = Design.empty.bind;
    Design.bindResource = Design.empty.bindResource;
    return Design;
}());
exports.Design = Design;
